package Logoved::DB::Stream::Listener::LogFileMangle;

use strict;
use warnings;
use Cwd;
use File::Basename;
use Exporter;
use parent 'Source::Shared::CLI', 'Exporter';
our $get_rpm_name_from_log;
our $get_rpm_nvr_from_log=\&nvr_from_filename;
our @EXPORT_OK=qw/$get_rpm_name_from_log $get_rpm_nvr_from_log/;

our $verbose=0;

my %KNOWN_MANGLE=(
    'default' => [ -mangler_sub => \&logname_mangle_default ],
    'dot' => [ -mangler_sub => \&logname_mangle_default ],
    'scope' => [ -mangler_sub => \&logname_mangle_scope ],
    'basename' => [ -mangler_sub => \&basename ],
    'dirname' => [ -mangler_sub => sub {basename(dirname($_[0]))} ],
    'autorepo' => [
	    -mangler_sub => \&logname_mangle_autorepo,
	    -set_rpm_getters => \&set_rpm_getters_autorepo,
	    -check_format_sub => \&check_logpath_format_autorepo,
    ],
    );


my %KNOWN_RPMGETTER=(
    'default' => \&set_rpm_getters_default,
    'autorepo' => \&set_rpm_getters_autorepo,
    );

my ($opt_mangle_logname,$opt_rpmname_method);
our $check_logpath_format_sub;

our @LONGOPT=(
    'logoved-mangle-logname=s' => \$opt_mangle_logname,
    'logoved-mangle-logname-dirname' => sub {$opt_mangle_logname='dirname'},
    'logoved-log2rpm=s' => \$opt_rpmname_method,
    );

sub print_longopt {
    print '  Logoved Report options
	--logoved-mangle-logname=('.join('|',keys(%KNOWN_MANGLE)).')
	--logoved-log2rpm=('.join('|',keys(%KNOWN_RPMGETTER)).')
';
}

sub process_options {
    die "LogFileMangle: unsupported mangle: ",$opt_mangle_logname
	if $opt_mangle_logname and not $KNOWN_MANGLE{$opt_mangle_logname};
    die "LogFileMangle: unsupported rpm name method: ",$opt_rpmname_method
	if $opt_rpmname_method and not $KNOWN_RPMGETTER{$opt_rpmname_method};
    $check_logpath_format_sub=\&check_logpath_format_autorepo if $opt_mangle_logname and $opt_mangle_logname eq 'autorepo';
}

my $file2reportname;
sub file2reportname { return $file2reportname }

sub calculate_mangled_filenames {
    my ($listeners_factory, $mangle_sub)=@_;
    my %_file_;
    foreach my $file (
	@{$listeners_factory->matched},
	@{$listeners_factory->not_matched},
	@{$listeners_factory->successful},
	) {
	$_file_{$file}=1;
    }
    my @mangler_names;
    if ($opt_mangle_logname) {
	@mangler_names=$opt_mangle_logname;
    } else {
	@mangler_names=qw/default dirname basename autorepo/;
    }
    my @manglers=map{__PACKAGE__->create($_)} @mangler_names;
    foreach my $file (keys(%_file_)) {
	foreach my $mangler (@manglers) {
	    $mangler->mangle($file);
	}
    }
    foreach my $mangler (@manglers) {
	next if $mangler->is_conflict;
	$file2reportname=$mangler->get_file2mangle;
	$mangler->set_rpm_getters if not $opt_rpmname_method;
    }
    if (not $file2reportname) {
	die "--logoved-mangle-logname=",join(',',@mangler_names),': input filenames does not fit or mangled names are not unique.',"\n";
    }
    &{$KNOWN_RPMGETTER{$opt_rpmname_method}} if $opt_rpmname_method;
}

sub create {
    my ($class, $name)=@_;
    my $opts=$KNOWN_MANGLE{$name};
    if ($opts) {
	return $class->new(-name => $name, @$opts);
    } else {
	die 'Oops! internal: unsupported name: ', $name;
    }
}

sub new {
    my $class=shift;
    my $self={
	-set_rpm_getters => \&set_rpm_getters_default,
	__known_mangle__ => {},
	@_,
    };
    die '-name is required' unless $self->{-name};
    die '-mangler_sub is required' unless $self->{-mangler_sub};
    bless $self, $class;
}

sub is_conflict {
    my ($self)=@_;
    return $self->{__known_name_conflict};
}

sub get_file2mangle {
    my ($self)=@_;
    return $self->{__file_2mangled_name};
}

sub set_rpm_getters {
    my ($self)=@_;
    return &{$self->{-set_rpm_getters}}();
}

sub mangle {
    my ($self,$file)=@_;
    my $__known_mangle__=$self->{__known_mangle__};
    my $mangled_name=&{$self->{-mangler_sub}}($file);
    $self->{__file_2mangled_name}->{$file}=$mangled_name;
    if ($__known_mangle__->{$mangled_name}) {
	$self->{__known_name_conflict}=1;
	print STDERR 'DEBUG: LogMangle: ',$self->{-name},': file=',$file,' mangled_name=',$mangled_name,"\n" if $verbose>3;
    } else {
	$__known_mangle__->{$mangled_name}=1;
    }
    my $check_format_sub=$self->{-check_format_sub};
    $self->{__known_name_conflict}=1 if $check_format_sub and not &$check_format_sub($file);
}

# ----------- filename mangles ----------

sub logname_mangle_default {
    return $_[0]=~s!/!.!gr;
}

sub logname_mangle_scope {
    return $_[0]=~s!/!::!gr;
}

sub logname_mangle_autorepo {
    return basename($_[0]=~s!/hsh\.log(\.[\w\d]+)$!$1!gr);
}

# ----------- get name for fixscripts ----------

sub name_from_dir {
    return basename(dirname($_[0]));
}

sub nvr_from_filename {
    my $nvr=basename($_[0]);
    $nvr=~s/\.log$//;
    $nvr=~s/\.(noarch|i[3456]86|pentimum\d|k6|athlon|athlon_xp|amd64|ia64|ia32e|alpha|arm|armv7l|e2k|aarch64|x86_64|mips64)$//;
    $nvr=~s/-\d+:/-/;
    return $nvr;
}

sub nvr_null {
    return '';
}

sub set_rpm_getters_default {
    $get_rpm_name_from_log=undef;
    $get_rpm_nvr_from_log=\&nvr_from_filename;
}

sub set_rpm_getters_autorepo {
    $get_rpm_name_from_log=\&name_from_dir;
    $get_rpm_nvr_from_log=\&nvr_null;
}

sub check_logpath_format_autorepo { $_[0] =~ m!TASK.*/[^/]+/hsh\.log\.[\w\d]+$!}

1;
