package Logoved::DB::Stream::Listener::Statistics;

use strict;
use warnings;
use Cwd;
use File::Basename;
use parent 'Source::Shared::CLI';
use Logoved::DB::Stream::Listener::LogFileMangle qw/$get_rpm_name_from_log $get_rpm_nvr_from_log/;
use Logoved::DB::Rule::Type qw/@errors_by_priority @fallbacks_by_priority @types_by_priority/;
use Logoved::DB::Rule::Resolver;
use Logoved::DB::Rule::Resolver::Plugin::Sourcedep;

our $verbose=0;

my $opt_flag_move_fixable;
my $opt_flag_resolve=1;
my $opt_flag_statistics=1;

our @LONGOPT=(
    'F|hide-fixable' => sub {$opt_flag_move_fixable=1},
    'show-fixable' => sub {$opt_flag_move_fixable=0},
    'resolve!' => \$opt_flag_resolve,
    'statistics!' => \$opt_flag_statistics,
    );

sub print_longopt {
    print '  Logoved Report options
	--hide-fixable
	--show-fixable
	--no-resolve
	--no-statistics
';
}

sub process_options {
    if ($opt_flag_resolve and ! -d $ENV{'HOME'}.'/.cache/distrodb/altlinux/sisyphus') {
	warn 'warning: dynamic DistroDB for altlinux:sisyphus not found.
	Option --resolve is disabled. Use --no-resolve or
	run /usr/bin/distrodb-update-repocop-db-altlinux-sisyphus
	to download latest DistroDB from repocop.altlinux.org.
';
	$opt_flag_resolve=0;
    }
}

my $by_report_type={
    'fixable' => {},
    'unresolved-fixes' => {},
    'error' => {},
    'overcome' => {},
    'warning' => {},
};
sub matches_by_type { return $by_report_type }

my $statistics_by_subpath={};
sub statistics_by_subpath { return $statistics_by_subpath }
my $resolved_statistics_by_subpath={};
sub resolved_statistics_by_subpath { return $resolved_statistics_by_subpath }

my $batch_fixscript='';
sub batch_fixscript { return $batch_fixscript }


my $by_ruletype_by_file={ map {$_=>{}} @types_by_priority };
sub handle_file_matches {
    my ($factory, $file, $matches)=@_;
    foreach my $match (@$matches) {
	my $file=$match->logfile;
	my $rule=$match->rule;
	my $rule_type=$rule->type;
	push @{$by_ruletype_by_file->{$rule_type}->{$file}}, $match;
    }
}

sub __from_by_file_to_by_rule {
    my ($origin_by_file, $dest_by_rule)=@_;
    foreach my $file (keys(%$origin_by_file)) {
	my $matches=$origin_by_file->{$file};
	foreach my $match (@$matches) {
	    my $rule_subpath=$match->rule->subpath;
	    push @{$dest_by_rule->{$rule_subpath}->{$file}}, $match;
	}
    }
}

sub calculate_report {
    my ($listeners_factory)=@_;
    my $subpath2rule=$listeners_factory->ruleset->_subpath2rule;

    #-------------------------------
    # sorting matches for report ...
    #-------------------------------
    &__from_by_file_to_by_rule($by_ruletype_by_file->{'warning'},$by_report_type->{'warning'});

    #-------------------------------
    # sorting matches for report ...
    # raise fallbacks to errors if errors not found
    #-------------------------------
    my $error_by_file={};
    my $error_by_rule=$by_report_type->{'error'};
    my $error_overcome_by_rule=$by_report_type->{'overcome'};
    foreach my $rule_type (@errors_by_priority, @fallbacks_by_priority) {
	my $fileset=$by_ruletype_by_file->{$rule_type};
	foreach my $file (keys(%$fileset)) {
	    if (not $error_by_file->{$file}) {
		my $matches=$fileset->{$file};
		$error_by_file->{$file}=$matches;
		foreach my $match (@$matches) {
		    my $rule=$match->rule;
		    my $rule_subpath=$rule->subpath;
		    my $status=$match->status;
		    $status//='error';
		    # ignore overcome fallback matches
		    next if $status eq 'success' and Logoved::DB::Rule::Type::is_fallback($rule_type);
		    if ($status eq 'error') {
			push @{$error_by_rule->{$rule_subpath}->{$file}}, $match;
		    } else {
			if ($rule->fix) {
			    # let it be fixed; in some cases it will help
			    push @{$error_by_rule->{$rule_subpath}->{$file}}, $match;
			}
			push @{$error_overcome_by_rule->{$rule_subpath}->{$file}}, $match;
		    }
		}
	    }
	}
    }

    #------------------------------

    my $matchset=$by_report_type->{'error'};

    #------------------------------
    # resolve matches
    #------------------------------
    if ($opt_flag_resolve) {
	my %resolver_cache;
	foreach my $rule_subpath (keys(%$matchset)) {
	    my $rule=$subpath2rule->{$rule_subpath};
	    my $resolver=$rule->resolver;
	    next unless defined $resolver;
	    my $filematches=$matchset->{$rule_subpath};
	    foreach my $file (keys(%$filematches)) {
		my $matches=$filematches->{$file};
		foreach my $match (@$matches) {
		    my $expanded_resolver_cmd=$resolver;
		    my $success=&cmdline_subst_groups($match,\$expanded_resolver_cmd);
		    if (!$success) {
			warn "Oops! resolver: bad group substitution in match:", $match->description;
		    } else {
			my $resolve=$resolver_cache{$expanded_resolver_cmd};
			#warn 'DEBUG:resolver_expanded=', $expanded_resolver_cmd,' cached=',(defined($resolve) ? $resolve:'null'),"\n" if $verbose>6;
			unless (defined($resolve)) {
			    my $status;
			    ($status,$resolve)=&Logoved::DB::Rule::Resolver::call_resolver($expanded_resolver_cmd);
			    warn 'DEBUG:call_resolver(', $expanded_resolver_cmd,'): status=',(defined($status) ? $status:'null'),' resolve=',(defined($resolve) ? @$resolve:'null'),"\n" if $verbose>6;
			    if ($status) {
				$resolver_cache{$expanded_resolver_cmd}=$resolve;
				die 'Oops! resolver: got undef for ',$expanded_resolver_cmd,' in match:', $match->description if not defined($resolve);
			    } else {
				warn 'Oops! resolver: failed ',$expanded_resolver_cmd,' in match:', $match->description;
			    }
			}
			#warn 'DEBUG:resolver: resolved=', @$resolve>0? @$resolve: 'empty', "\n";
			$match->resolved_value($resolve);
		    }
		}
	    }
	}
    }
    #------------------------------
    # count statistics
    #------------------------------
    if ($opt_flag_resolve) {
	foreach my $rule_subpath (keys(%$matchset)) {
	    my $rule=$subpath2rule->{$rule_subpath};
	    my $statistics=$rule->statistics;
	    next unless defined $statistics;
	    my ($statistics_lambda,@lambda_args)=&get_statistics_lambda($statistics);
	    #warn "lambda_args=@lambda_args\n";
	    my %count_resolved;
	    my %count_unresolved;
	    my $filematches=$matchset->{$rule_subpath};
	    foreach my $file (keys(%$filematches)) {
		my $matches=$filematches->{$file};
		$#{$matches}=0 if $#{$matches}>0 and $rule->use_first_match;
		my %unique_unresolved_matches;
		my %unique_resolved_matches;
		foreach my $match (@$matches) {
		    my $statdata;
		    # apt/E-broken.logoved; multiple patterns, only 1 has match
		    eval {
			$statdata=&$statistics_lambda($match, @lambda_args);
		    };
		    if ($@) {
			# warning? or set rule flag? not to clog stderr
			next;
		    }
		    my $resolved=$match->resolved_value;
		    if (!$resolved or 0==@$resolved) {
			$unique_unresolved_matches{$statdata}=1;
		    } else {
			# TODO: add resolved value - and what sort it is? by resolved value first!
			$unique_resolved_matches{$statdata.' -> '.join('|',@$resolved)}=1;
		    }
		}
		map {$count_unresolved{$_}++} keys(%unique_unresolved_matches);
		map {$count_resolved{$_}++} keys(%unique_resolved_matches);
	    }
	    my $statistics_header=$rule->statistics_header;
	    my @sorted_items=sort {$count_unresolved{$b} <=> $count_unresolved{$a}} keys(%count_unresolved);
	    $statistics_by_subpath->{$rule_subpath}=
		($statistics_header ? join("\n", @$statistics_header)."\n" :'').
		join("\n", map {sprintf('%3i: %s',$count_unresolved{$_}, $_)} @sorted_items)."\n"
		if @sorted_items;
	    #print $statistics_by_subpath->{$rule_subpath};
	    @sorted_items=sort {$count_resolved{$b} <=> $count_resolved{$a}} keys(%count_resolved);
	    $resolved_statistics_by_subpath->{$rule_subpath}=
		($statistics_header ? join("\n", @$statistics_header)."\n" :'').
		join("\n", map {sprintf('%3i: %s',$count_resolved{$_}, $_)} @sorted_items)."\n"
		if @sorted_items;
	    #print $resolved_statistics_by_subpath->{$rule_subpath};
	}
    }

    #------------------------------
    # find fixable
    #------------------------------
    my $is_fixable_at_rule_subpath_file;
    foreach my $rule_subpath (keys(%$matchset)) {
	my $rule=$subpath2rule->{$rule_subpath};
	my $filesref=$matchset->{$rule_subpath};
	my $rule_fix=$rule->fix;
	next unless $rule_fix;
	my $fix_if_resolved=$rule->fix_if_resolved;
	foreach my $file (keys(%{$filesref})) {
	    my $this_file_fix=$rule_fix;
	    # loghook hack
	    if ($rule_fix=~/^loghook\b/) {
		my $abs_file_path=Cwd::abs_path($file);
		$this_file_fix=~s/^loghook\b/hook --env 'LOGOVED_LOGFILE=$abs_file_path'/;
	    }
	    foreach my $match (@{$filesref->{$file}}) {
		my $resolved=$match->resolved_value();
		next if $fix_if_resolved and (not $resolved or 0==@$resolved);
		$is_fixable_at_rule_subpath_file->{$rule_subpath}->{$file}=1;
		my $expanded_fix=$this_file_fix;
		my $success=&cmdline_subst_groups($match,\$expanded_fix);
		if (!$success) {
		    warn "Oops! fix: bad group substitution in match:", $match->description;
		} else {
		    my $rpm_nvr=&$get_rpm_nvr_from_log($file);
		    my $rpm_name=$rpm_nvr;
		    if ($get_rpm_name_from_log and not $rpm_nvr) {
			$rpm_name=&$get_rpm_name_from_log($file);
		    } else {
			$rpm_name=~s/-[^-]+-[^-]+$//;
		    }
		    #warn "FIX1: file=$file rpm_nvr=$rpm_nvr rpm_name=$rpm_name\n";
		    $batch_fixscript .= $rpm_name."\t".$rpm_nvr."\t".$expanded_fix."\n";
		}
	    }
	}
    }

    #------------------------------
    # move out fixable from the report's error tree
    #------------------------------
    if ($opt_flag_move_fixable) {
	#my $matchset=$by_report_type->{'error'};
	my $fixset={};
	foreach my $rule_subpath (keys(%$matchset)) {
	    my $is_fixable_file=$is_fixable_at_rule_subpath_file->{$rule_subpath};
	    my $fileset=$matchset->{$rule_subpath};
	    foreach my $file (keys(%$fileset)) {
		if ($is_fixable_file->{$file}) {
		    $fixset->{$rule_subpath}->{$file}=$fileset->{$file};
		    delete $fileset->{$file} if $opt_flag_move_fixable;
		}
	    }
	}
	$by_report_type->{'fixable'}=$fixset;
    }
}

sub get_statistics_lambda {
    my $statistics=shift;
    if ($statistics=~/^\%\{GROUP:(\d+)\}/) {
	my $group_no=$1;
	return sub {'ERROR: unsupported GROUP #0: '.$statistics} if $group_no==0;
	# also good, but longer
	#return (sub {$_[0]->groups->[$_[1]]}, $group_no-1);
	return sub {my $groups=$_[0]->groups; die "statistics: no match in ",$_[0]->description if not ref $groups; $groups->[$group_no-1]};
    } elsif ($statistics=~/^\%\{MULTILINE_BEGIN_GROUP:(\d+)\}/) {
	my $group_no=$1;
	return sub {'ERROR: unsupported MULTILINE_BEGIN_GROUP #0: '.$statistics} if $group_no==0;
	return sub {$_[0]->multiline_begin_groups->[$group_no-1]};
    } else {
	return sub {'ERROR: unsupported statistics: '.$statistics};
    }
}

sub cmdline_subst_groups {
    my ($match, $fixref)=@_;
    my $success=1;
    my $group_subst=sub {
	my $keyword=shift;
	my $num=shift;
	my $groups;
	if ($keyword eq 'GROUP') {
	    $groups=$match->groups;
	} elsif ($keyword eq 'MULTILINE_BEGIN_GROUP') {
	    $groups=$match->multiline_begin_groups;
	} elsif ($keyword eq 'MULTILINE_END_GROUP') {
	    $groups=$match->multiline_end_groups;
	} else {
	    die "Unsupported group: ", $keyword, ' in ',$$fixref;
	}
	if ($groups and $#{$groups}>=$num-1 and $num>=1) {
	    return $groups->[$num-1];
	} else {
	    $success=0;
	    warn 'Bad fix in rule ', $match->rule->rulefile, ': %{',$keyword,':',$num,'} not found',"\n";
	}
	return '%{'.$keyword.':'.$num.'}';
    };
    $$fixref=~s/\%\{(MULTILINE_BEGIN_GROUP|MULTILINE_END_GROUP|GROUP):(\d+)\}/&$group_subst($1,$2)/ge;
    $$fixref=~s/\%\{MULTILINE_BODY_GROUP:\*\}/join(' ',@{$match->multiline_body_groups()})/ge;
    return $success;
}

1;
